<?php

namespace Drupal\entrypoints\Plugin;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\entrypoints\EntrypointDefinition;

/**
 * The interface implemented by all entrypoints input handler plugins.
 */
interface EntrypointsInputHandlerPluginInterface extends PluginInspectionInterface {

  /**
   * Prepares server-side rendering with the given input.
   *
   * This method will be called as the very first step, in order to prepare the
   * input for every entrypoint. You can adjust the input, and also may change
   * the list of entrypoints that are to be rendered. You could also change the
   * renderer, but that is not recommended. It should be noted that this plugin
   * would never be invoked, if there are no configured SSR entrypoints.
   *
   * @param array &$input
   *   An array that contains input, which will be passed to the rendering
   *   process as "DrupalServer" Javascript object. The values can be changed
   *   and extended by further data that might be needed within the SSR process.
   * @param \Drupal\entrypoints\EntrypointDefinition[] &$definitions
   *   All entrypoint definitions that are to be rendered on the server-side.
   * @param \Drupal\entrypoints\Plugin\EntrypointsRendererPluginInterface &$renderer
   *   The renderer plugin that will be used for rendering.
   * @param \Drupal\Core\Render\HtmlResponse|mixed &$context
   *   Further parameter that might be relevant. For the response build chain,
   *   this parameter is the current response object.
   */
  public function prepareRender(array &$input, array &$definitions, EntrypointsRendererPluginInterface &$renderer, &$context = NULL);

  /**
   * Invokes right before the given entrypoint is being server-side rendered.
   *
   * This enables you to make changes and additions to the input array that will
   * be passed to the given entrypoint. It should be noted that the input
   * additions and changes would also affect subsequent entrypoints that are to
   * be rendered, because the input for subsequent entrypoints is the output
   * generated by the given entrypoint (i.e. each SSR entrypoint generates the
   * input for the next SSR entrypoint to be rendered). If you want to hide
   * information or prevent possible conflicts and overheads, you should also
   * implement ::afterRender() to take necessary steps.
   *
   * @param array &$input
   *   An array that contains input, which will be passed to the rendering
   *   process as "DrupalServer" Javascript object. The values can be changed
   *   and extended by further data that might be needed within the SSR process.
   * @param \Drupal\entrypoints\EntrypointDefinition $definition
   *   The entrypoint that is about to be server-side rendered.
   * @param \Drupal\entrypoints\Plugin\EntrypointsRendererPluginInterface $renderer
   *   The renderer plugin that will be used for rendering.
   * @param \Drupal\Core\Render\HtmlResponse|mixed &$context
   *   Further parameter that might be relevant. For the response build chain,
   *   this parameter is the current response object.
   */
  public function beforeRender(array &$input, EntrypointDefinition $definition, EntrypointsRendererPluginInterface $renderer, &$context = NULL);

  /**
   * Invokes right after the given entrypoint was server-side rendered.
   *
   * @param array $input
   *   Contains the input that was passed to the rendering process. It should
   *   be noted that this value is not relevant for the subsequent entrypoints,
   *   instead it is the output that will be used as input for subsequent
   *   rendering processes.
   * @param array|false &$output
   *   The output that was generated by the rendering process. Typically
   *   contains updated values regards the given $input argument. The value is
   *   FALSE if the renderer execution did not deliver any result, but is not
   *   considered to be an error and thus is not worth throwing an exception.
   * @param \Drupal\entrypoints\EntrypointDefinition $definition
   *   The entrypoint that just got rendered.
   * @param \Drupal\entrypoints\Plugin\EntrypointsRendererPluginInterface $renderer
   *   The renderer plugin that was used for rendering.
   * @param \Drupal\Core\Render\HtmlResponse|mixed &$context
   *   Further parameter that might be relevant. For the response build chain,
   *   this parameter is the current response object.
   */
  public function afterRender(array $input, &$output, EntrypointDefinition $definition, EntrypointsRendererPluginInterface $renderer, &$context = NULL);

  /**
   * Finishing steps after all entrypoints got server-side rendered.
   *
   * @param array &$output
   *   An array that contains generated output. This argument is not being used
   *   from the outside anymore, but it could be used for passing information
   *   between input handler plugins within their finishing routines.
   * @param \Drupal\entrypoints\EntrypointDefinition[] &$definitions
   *   All entrypoint definitions that got rendered on the server-side.
   * @param \Drupal\entrypoints\Plugin\EntrypointsRendererPluginInterface &$renderer
   *   The renderer plugin that was used for rendering.
   * @param \Drupal\Core\Render\HtmlResponse|mixed &$context
   *   Further parameter that might be relevant. For the response build chain,
   *   this parameter is the current response object.
   */
  public function finishRender(array &$output, array &$definitions, EntrypointsRendererPluginInterface &$renderer, &$context = NULL);

}
